import datetime
import urllib2

from BeautifulSoup import BeautifulStoneSoup

from rpi_courses.web import get
from rpi_courses.parser.features import *  # all object postfixed with '_feature' will get used.


class CourseCatalog(object):
    """Represents the RPI course catalog.

    This takes a BeautifulSoup instance (usually a BeautifulStoneSoup instance)
    allows an object-oriented method of accessing the data.
    """

    # this keeps the parsing separate from the actual data fetching.
    # We'll call each feature we imported that ends with '_feature'.
    FEATURES = [obj for name, obj in globals().iteritems() if name.endswith('_feature')]

    def __init__(self, soup=None):
        """Instanciates a CourseCatalog given a BeautifulSoup instance.
        Pass nothing to initiate an empty course catalog.
        """
        if soup is not None:
            self.parse(soup)

    @staticmethod
    def from_string(xml_str):
        "Creates a new CourseCatalog instance from an string containing xml."
        return CourseCatalog(BeautifulStoneSoup(xml_str,
            convertEntities=BeautifulStoneSoup.XML_ENTITIES
        ))

    @staticmethod
    def from_stream(stream):
        "Creates a new CourseCatalog instance from a filehandle-like stream."
        return CourseCatalog.from_string(stream.read())

    @staticmethod
    def from_file(filepath):
        "Creates a new CourseCatalog instance from a local filepath."
        with open(filepath) as f:
            return CourseCatalog.from_stream(f)

    @staticmethod
    def from_url(url):
        "Creates a new CourseCatalog instance from a given url."
        return CourseCatalog.from_string(get(url))

    def parse(self, soup):
        "Parses the soup instance as RPI's XML course catalog file."
        for feature in self.FEATURES:
            feature(self, soup)

    def crosslisted_with(self, crn):
        """Returns all the CRN courses crosslisted with the given crn.
        The returned crosslisting does not include the original CRN.
        """
        return tuple([c for c in self.crosslistings[crn].crns if c != crn])

    def find_courses(self, partial):
        """Finds all courses by a given substring. This is case-insensitive.
        """
        partial = partial.lower()
        keys = self.courses.keys()
        keys = [k for k in keys if k.lower().find(partial) != -1]
        courses = [self.courses[k] for k in keys]
        return list(set(courses))

    def get_courses(self):
        """Returns all course objects from this catalog.
        """
        return self.courses.values()

    def find_course_by_crn(self, crn):
        """Searches all courses by CRNs. Not particularly efficient.
        Returns None if not found.
        """
        for name, course in self.courses.iteritems():
            if crn in course:
                return course
        return None

    def find_course_and_crosslistings(self, partial):
        """Returns the given course and all other courses it is
        crosslisted with.
        """
        course = self.find_course(partial)
        crosslisted = self.crosslisted_with(course.crn)
        return (course,) + tuple(map(self.find_course_by_crn, crosslisted))



import logging
import logging.handlers
import sys

DEBUG = False
LOG_FILENAME = 'logging'

logger = logging.getLogger('rpi_courses')


if hasattr(logging, 'NullHandler'):
    NullHandler = logging.NullHandler
else:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

# fallback, so there's no warning of no handlers
logger.addHandler(NullHandler())

SIS_URL = "http://sis.rpi.edu/reg/"
# where ROCS's xml files are located.
ROCS_URL = "http://sis.rpi.edu/reg/rocs/"

# TODO: this may be good to parse (for textbooks)
HTML_URL = "http://sis.rpi.edu/stuclshr.htm"

COMM_URL = "http://www.rpi.edu/dept/srfs/CI"

# it would be best to get this data from a reliable data source:
# like the course catalog: http://catalog.rpi.edu/content.php?catoid=10&navoid=232
# instead of manually entering this data.

# code: name
DEPARTMENTS = dict(
    ADMN="Administration",
    ARCH="Architecture",
    ARTS="Arts",
    ASTR="Astronomy",
    BCBP="Biochemistry and Biophysics",
    BIOL="Biology",
    BMED="Biomedical Engineering",
    CHEM="Chemistry",
    CHME="Chemical Engineering",
    CISH="Computer Science at Hartford",
    CIVL="Civil Engineering",
    COGS="Cognitive Science",
    COOP="Cooperative Education",
    COMM="Communication",
    CSCI="Computer Science",
    ECON="Economics",
    ECSE="Electrical, Computer, and Systems Engineering",
    ENGR="General Engineering",
    ENVE="Environmental Engineering",
    EPOW="Electrical, Computer, and Systems Engineering",
    ERTH="Earth and Environmental Sciences",
    ESCI="Engineering Science",
    EXCH="Exchange (Study Abroad)",
    IENV="Interdisciplinary Environmental",
    IHSS="Interdisciplinary Studies",
    ISCI="Interdisciplinary Science",
    ISYE="Industrial and Systems Engineering",
    ITWS="Information Technology and Web Science",
    LANG="Foreign Languages",
    LGHT="Lighting",
    LITR="Literature",
    MANE="Mechanical, Aerospace, and Nuclear Engineering",
    MATH="Mathematics",
    MATP="Mathematical Programming, Probability, and Statistics",
    MGMT="Management",
    MTLE="Materials Science and Engineering",
    NSST="Natural Science for School Teachers",
    PHIL="Philosophy",
    PHYS="Physics",
    PSYC="Psychology",
    STSH="Science and Technology Studies (Humanities Courses)",
    STSS="Science and Technology Studies (Social Sciences Courses)",
    USAF="Aerospace Studies (Air Force ROTC)",
    USAR="Military Science (Army ROTC)",
    USNA="Naval Science (Navy ROTC)",
    WRIT="Writing",
)



import collections

from config import logger


def safeInt(n, warn_only=False):
    """Throws an exception if the number starts with a 0 (may be significant).

    If the value cannot be converted to an int, it is returned as is.
    """
    if str(n).startswith('0'):
        if not warn_only:
            raise TypeError("Unsafe Int: " + str(n))
        print "Unsafe Int: %s" % n
        return int(n)
    try:
        return int(n)
    except ValueError:
        return n


# from SO: http://stackoverflow.com/questions/2703599/what-would-be-a-frozen-dict
class FrozenDict(collections.Mapping):
    """Defines an immutable dict type."""

    FROZEN_TYPES = {
        set: frozenset,
        list: tuple,
    }

    def __init__(self, *args, **kwargs):
        self._hash = None
        self._d = {}
        for key, vals in dict(*args, **kwargs).items():
            self._d[self._freeze(key)] = self._freeze(vals)

    def _freeze(self, value):
        return self.FROZEN_TYPES.get(type(value), lambda x: x)(value)

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getitem__(self, key):
        return self._d[key]

    def __setitem__(self, key, value):
        raise TypeError("FrozenDict is immutable.")

    def __repr__(self):
        return "FrozenDict(%r)" % self._d

    def __hash__(self):
        if self._hash is None:
            self._hash = 0
            for key, value in self.iteritems():
                self._hash ^= hash(key)
                self._hash ^= hash(value)
        return self._hash

FrozenDict.FROZEN_TYPES[dict] = FrozenDict



import unittest


class TestCaseForModel(unittest.TestCase):
    def assertPeriodEquals(self, period1, period2):
        """
        Manually compares each property of two period instances.
        """
        self.assertEquals(period1.type, period2.type)
        self.assertEquals(period1.instructor, period2.instructor)
        self.assertEquals(period1.location, period2.location)
        self.assertEquals(period1.start, period2.start)
        self.assertEquals(period1.end, period2.end)
        self.assertEquals(period1.int_days, period2.int_days)

    def assertIn(self, item, collection):
        assert item in collection

    def assertPeriodEqual(self, period1, period2):
        self.assertPeriodEquals(period1, period2)

    def assertPeriodsEqual(self, periods1, periods2):
        """
        Invokes self.assertSectionEquals to two tuples of periods.
        """
        self.assertEquals(len(periods1), len(periods2))
        for i in range(len(periods1)):
            self.assertPeriodEquals(periods1[i], periods2[i])

    def assertPeriodsEquals(self, periods1, periods2):
        self.assertPeriodsEqual(period1, period2)

    def assertSectionEquals(self, section1, section2):
        """
        Manually compares each property of two section instances.
        """
        self.assertEquals(section1.crn, section2.crn)
        self.assertEquals(section1.num, section2.num)
        self.assertEquals(section1.seats_taken, section2.seats_taken)
        self.assertEquals(section1.seats_total, section2.seats_total)
        self.assertEquals(section1.notes, section2.notes)
        self.assertPeriodsEqual(section1.periods, section2.periods)

    def assertSectionEqual(self, section1, section2):
        self.assertSectionEquals(section1, section2)

    def assertSectionsEquals(self, sections1, sections2):
        """
        Invokes self.assertSectionEquals to two tuples of sections.
        """
        self.assertEquals(len(sections1), len(sections2))
        for i in range(len(sections1)):
            self.assertSectionEquals(sections1[i], sections2[i])

    def assertSectionsEqual(self, sections1, sections2):
        self.assertSectionsEquals(sections1, sections2)

    def assertCourseEquals(self, course1, course2):
        """
        Manually compares each property of two course instances.
        """
        self.assertEquals(course1.name, course2.name)
        self.assertEquals(course1.dept, course2.dept)
        self.assertEquals(course1.num, course2.num)
        self.assertEquals(course1.cred, course2.cred)
        # technically, we should compare grade_type
        # but we only want to know if the course is pass/fail anyway
        self.assertEquals(course1.is_pass_or_fail, course2.is_pass_or_fail)
        self.assertSectionsEqual(course1.sections, course2.sections)

    def assertCourseEqual(self, course1, course2):
        self.assertCourseEquals(course1, course2)



"""models.py - The classes that store the data retrieved by features.py

Generally, all instances should be read only.
"""
from datetime import time

from rpi_courses.utils import safeInt, FrozenDict
from rpi_courses.config import DEPARTMENTS


class ReadOnly(object):
    """All attributes that are prefixed with a single underscore will have
    a equivalent getter property without the underscore prefix.

    This restricts all those attributes to read-only.

    eg - setting self._foo, allows you to access it publicly using self.foo
    """
    def __getattr__(self, key):
        if not key.startswith('_') and not key.endswith('__'):
            value = getattr(self, '_' + key)

            the_type = type(value)
            if the_type == list:
                return tuple(value)
            if the_type == dict and type_type != FrozenDict:
                return FrozenDict(the_type)

            return value
        raise AttributeError("type object %r has no attribute %r" % (
            self.__class__.__name__, key
        ))


class CrossListing(ReadOnly):
    """Represents a crosslisted set of CRNs and Seats.
    This is immutable once created.
    """
    def __init__(self, crns, seats):
        self._crns, self._seats = frozenset(crns), seats

    def __eq__(self, other):
        return self.crns == other.crns and self.seats == other.seats

    def __add__(self, other_crosslisting):
        return self.__class__(self._crns + other_crosslisting._crns, self._seats + other_crosslisting._seats)


DAY_MAPPER = {
    0: 'Monday',
    1: 'Tuesday',
    2: 'Wednesday',
    3: 'Thursday',
    4: 'Friday',
    5: 'Saturday',
    6: 'Sunday',
}


class Period(ReadOnly):
    def __init__(self, type, instructor, start, end, location, int_days):
        self._type, self._instructor, self._location = \
            type.strip(), instructor.strip(), location.strip()

        try:
            self._start, self._end = int(start), int(end)
        except ValueError:
            # == '** TBA **'
            self._start = self._end = None
        self._int_days = tuple(map(int, int_days))
        self.__hash = None

    def __repr__(self):
        return "<Period: start=%r end=%r days=%r>" % (
            self.start, self.end, self.days
        )

    def __eq__(self, other):
        return (self.type, self.instructor, self.location, self.start, self.end, \
            self.int_days) == (other.type, other.instructor, other.location, \
            other.start, other.end, other.int_days)

    def __hash__(self):
        if self.__hash is None:
            result = hash(self.type)
            for v in (self.instructor, self.location, self.start, self.end, self.int_days):
                result = result ^ hash(v)
            self.__hash = result
        return self.__hash

    @staticmethod
    def from_soup_tag(tag):
        "Returns a new Period instance from the given beautifulsoup tag."
        days = []
        for elem in tag.findAll(recursive=False):
            if elem.name != 'day':
                raise TypeError("Unknown tag found: " + str(elem))
            days.append(elem.string)
        return Period(
            tag['type'], tag['instructor'], tag['start'], tag['end'],
            tag['location'], days
        )

    @property
    def time_range(self):
        return (self.start, self.end)

    @property
    def start_time(self):
        if self.start is None:
            return None
        s = str(self.start)
        hours, minutes = (s[:-2], s[-2:])
        return time(hour=int(hours), minute=int(minutes))

    @property
    def end_time(self):
        if self.end is None:
            return None
        s = str(self.end)
        hours, minutes = (s[:-2], s[-2:])
        return time(hour=int(hours), minute=int(minutes))

    def conflicts_with(self, period):
        "Checks this period conflicts with another period."
        if self.tba or period.tba:
            return False
        same_day = False
        for i in self.int_days:
            if i in period.int_days:
                same_day = True

        if not same_day:
            return False

        return self.start <= period.start <= self.end or \
            period.start <= self.start <= period.end or \
            self.start <= period.end <= self.end or \
            period.start <= self.end <= period.end

    @property
    def tba(self):
        "The time period hasn't been announced yet."
        return self.start is None or self.end is None

    @property
    def is_lecture(self):
        return self.type == 'LEC'

    @property
    def is_studio(self):
        return self.type == 'STU'

    @property
    def is_lab(self):
        return self.type == 'LAB'

    @property
    def is_testing_period(self):
        return self.type == 'TES'

    @property
    def is_recitation(self):
        return self.type == 'REC'

    @property
    def days(self):
        return tuple(map(DAY_MAPPER.get, self.int_days))


class Section(ReadOnly):
    """A section is a particular timeslot to take the given course.
    It is uniquely represented in SIS via CRN. The CRN is used for
    registration.
    """
    def __init__(self, crn, num, taken, total, periods, notes):
        self._crn, self._seats_taken, self._seats_total = \
            safeInt(crn), int(taken), int(total)
        self._num = num
        self._periods = tuple(periods)
        self._notes = tuple(set(notes))
        self.__hash = None

    #def __eq__(self, other):
    #    return (self.crn, self.num, self.seats_taken, self.seats_total, \
    #        self.periods, self.notes) == (other.crn, other.num, \
    #        other.seats_taken, other.periods, self.notes)

    def __hash__(self):
        if self.__hash is None:
            result = hash(self.crn)
            for v in (self.seats_taken, self.seats_total, self.num, self.periods, self.notes):
                result = result ^ hash(v)
            self.__hash = result
        return self.__hash

    def conflicts_with(self, section):
        for period in self.periods:
            for p in section.periods:
                if period.conflicts_with(p):
                    return True
        return False

    @staticmethod
    def from_soup_tag(tag):
        "Returns an instance from a given soup tag."
        periods = []
        notes = []
        for elem in tag.findAll(recursive=False):
            if elem.name not in ('period', 'note'):
                raise TypeError("Unknown tag found: " + str(elem))
            if elem.name == 'note':
                notes.append(elem.string.strip())
            elif elem.name == 'period':
                periods.append(Period.from_soup_tag(elem))
        return Section(
            tag['crn'], tag['num'], tag['students'], tag['seats'],
            periods, notes
        )

    @property
    def is_study_abroad(self):
        return self.num in ('SA', 'EXC')

    @property
    def is_off_campus(self):
        return self.num.startswith('OC')

    @property
    def is_valid(self):
        """Returns True if the given data for this section is valid."""
        return self.seats_total > 0

    @property
    def is_filled(self):
        "Returns True if the course is full and not accepting any more seats."
        return self.seats_taken >= self.seats_total

    @property
    def seats_left(self):
        """Returns the number of seats left.
        A negative number indicates more seats taken than are available.
        """
        return self.seats_total - self.seats_taken

    def __repr__(self):
        return "<Section: crn=%(crn)r num=%(num)r seats=%(used)r/%(total)r>" % {
            'crn': self.crn,
            'num': self.num,
            'used': self.seats_taken,
            'total': self.seats_total,
        }

    def __eq__(self, other):
        if isinstance(other, Section):
            return (
                self.crn == other.crn and self.num == other.num and
                self.periods == other.periods
            )
        return False


class Course(ReadOnly):
    """Represents a particular kind of course and its sections.
    Also immutable once created.
    """
    def __init__(self, name, dept, num, credmin, credmax, grade_type, sections):
        self._name, self._dept, self._num, self._cred, self._grade_type = \
            name.strip(), dept.strip(), safeInt(num, warn_only=True), \
            (int(credmin), int(credmax)), grade_type.strip(),
        self._sections = tuple(sections)
        self.__free_sections = None
        self.__hash = None

    def __contains__(self, crn):
        for section in self.sections:
            if section.crn == crn:
                return True
        return False

    def __eq__(self, other):
        return (
            self.num == other.num and self.cred == other.cred and
            self.grade_type == other.grade_type and
            self.name == other.name and self.dept == other.dept and
            self.sections == other.sections
        )

    def __hash__(self):
        if self.__hash is None:
            result = hash(self.name)
            for v in (self.dept, self.num, self.cred[0], self.cred[1], self.grade_type, self.sections):
                result = result ^ hash(v)
            self.__hash = result
        return self.__hash

    def __str__(self):
        return "%(name)s %(dept)s %(num)s" % {
            'name': self.name,
            'dept': self.dept,
            'num': self.num,
        }

    def __repr__(self):
        return "<Course: %(name)r, %(dept)r, %(num)r, %(mincred)r, %(maxcred)r, %(grade_type)r, section_count=%(section_count)s>" % {
            'name': self.name,
            'dept': self.dept,
            'num': self.num,
            'mincred': self.cred[0],
            'maxcred': self.cred[1],
            'grade_type': self.grade_type,
            'section_count': len(self.sections)
        }

    @property
    def full_dept(self):
        "Uses DEPARTMENT to convert the shorthand name into a full name of the department."
        return DEPARTMENTS.get(self.dept)

    @property
    def available_sections(self):
        if self.__free_sections is None:
            self.__free_sections = tuple(s for s in self.sections if s.seats_left > 0)
        return self.__free_sections

    @property
    def credits(self):
        """Returns either a tuple representing the credit range or a
        single integer if the range is set to one value.

        Use self.cred to always get the tuple.
        """
        if self.cred[0] == self.cred[1]:
            return self.cred[0]
        return self.cred

    @property
    def is_pass_or_fail(self):
        "Returns True if this course is graded as pass/fail."
        return self.grade_type.lower() == 'satisfactory/unsatisfactory'

    @property
    def code(self):
        "Returns the 'dept num'."
        return u"%s %s" % (self.dept, self.num)

    @staticmethod
    def from_soup_tag(tag):
        "Creates an instance from a given soup tag."
        sections = [Section.from_soup_tag(s) for s in tag.findAll('section')]
        return Course(
            tag['name'], tag['dept'], int(tag['num']), tag['credmin'],
            tag['credmax'], tag['gradetype'], [s for s in sections if s.is_valid]
        )



"""features.py - Implements all parsing of the XML file.

All functions related to parsing the XML file are here. To be
automatically imported by the CourseCatalog class, postfix the
function name with '_feature'
"""
import datetime

from rpi_courses.utils import FrozenDict, safeInt
from rpi_courses.config import logger, DEBUG
from rpi_courses.models import CrossListing, Course


def timestamp_feature(catalog, soup):
    """The datetime the xml file was last modified.
    """
    catalog.timestamp = int(soup.coursedb['timestamp'])
    catalog.datetime = datetime.datetime.fromtimestamp(catalog.timestamp)
    logger.info('Catalog last updated on %s' % catalog.datetime)


def semester_feature(catalog, soup):
    """The year and semester information that this xml file hold courses for.
    """
    raw = soup.coursedb['semesternumber']
    catalog.year = int(raw[:4])

    month_mapping = {1: 'Spring', 5: 'Summer', 9: 'Fall'}
    catalog.month = int(raw[4:])
    catalog.semester = month_mapping[catalog.month]

    catalog.name = soup.coursedb['semesterdesc']

    logger.info('Catalog type: %s' % catalog.name)


def crosslisting_feature(catalog, soup):
    """Parses all the crosslistings. These refer to the similar CRNs,
    such as a grad & undergrad level course.
    """
    listing = {}
    for elem in soup.coursedb.findAll('crosslisting'):
        seats = int(elem['seats'])
        crns = [safeInt(crn.string) for crn in elem.findAll('crn')]

        # we want to refer to the same object to save space
        cl = CrossListing(crns, seats)
        for crn in crns:
            listing[crn] = cl
    catalog.crosslistings = FrozenDict(listing)

    logger.info('Catalog has %d course crosslistings' % len(catalog.crosslistings))


def course_feature(catalog, soup):
    """Parses all the courses (AKA, the most important part).
    """
    courses = {}
    course_crns = {}
    for course in soup.findAll('course'):
        c = Course.from_soup_tag(course)
        courses[str(c)] = c
    catalog.courses = courses
    catalog.courses
    logger.info('Catalog has %d courses' % len(courses))



import os

ROOT = os.path.abspath(os.path.dirname(__file__))
ROOT_TEST_DATA = os.path.join(ROOT, 'test_data')

XML_FILE = os.path.join(ROOT_TEST_DATA, 'rpi_courses.xml')
XML_PARSER_FILE = os.path.join(ROOT_TEST_DATA, 'rpi_shorthand.xml')
XML_SCHEDULE_TEST_FILE = os.path.join(ROOT_TEST_DATA, 'rpi_schedule_test.xml')
XML_SCHEDULE_TEST_CONFLICT_FILE = os.path.join(ROOT_TEST_DATA, 'rpi_schedule_conflict_test.xml')
SIS_FILE_SPRING2012 = os.path.join(ROOT_TEST_DATA, 'sis_courses_spring2012.htm')

HTML = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<HTML>
 <HEAD>
  <TITLE>Index of /reg/rocs</TITLE>
 </HEAD>
 <BODY>
<H1>Index of /reg/rocs</H1>
<PRE><IMG SRC="/icons/blank.gif" ALT="     "> <A HREF="?N=D">Name</A>                    <A HREF="?M=A">Last modified</A>       <A HREF="?S=A">Size</A>  <A HREF="?D=A">Description</A>
<HR>
<IMG SRC="/icons/back.gif" ALT="[DIR]"> <A HREF="/reg/">Parent Directory</A>        18-Oct-2010 20:37      -
<IMG SRC="/icons/unknown.gif" ALT="[   ]"> <A HREF="201001.xml">201001.xml</A>              28-Sep-2010 10:23   719k
<IMG SRC="/icons/unknown.gif" ALT="[   ]"> <A HREF="201005.xml">201005.xml</A>              28-Sep-2010 10:23    42k
<IMG SRC="/icons/unknown.gif" ALT="[   ]"> <A HREF="201009.xml">201009.xml</A>              03-Nov-2010 21:00   730k
<IMG SRC="/icons/text.gif" ALT="[TXT]"> <A HREF="201009.xml_old">201009.xml_old</A>          28-Sep-2010 10:23   698k
<IMG SRC="/icons/unknown.gif" ALT="[   ]"> <A HREF="201101.xml">201101.xml</A>              17-Nov-2010 14:16   713k
<IMG SRC="/icons/text.gif" ALT="[TXT]"> <A HREF="testrocs.txt">testrocs.txt</A>            28-Sep-2010 10:23     1k
</PRE><HR>
<ADDRESS>Oracle-Application-Server-10g/10.1.2.2.0 Oracle-HTTP-Server Server at sis.rpi.edu Port 80</ADDRESS>
</BODY></HTML>"""



from rpi_courses.sis_parser import CourseCatalog
from rpi_courses import models
from test.rpi_courses.utils import TestCaseForModel

from mock import Mock

import os
import unittest
from time import mktime
from constants import SIS_FILE_SPRING2012

with open(SIS_FILE_SPRING2012) as f:
    CONTENTS_SPRING2012 = f.read()
catalog_spring2012 = None


class TestCatalogSpring2012(TestCaseForModel):
    def setUp(self):
        global catalog_spring2012
        if catalog_spring2012 is None:
            catalog_spring2012 = CourseCatalog.from_string(CONTENTS_SPRING2012)
        self.catalog = catalog_spring2012

#    def test_crosslisting(self):
#        crns = self.catalog.crosslisted_with(76093)
#        for e in (76097, 76098, 76099):
#            self.assertIn(e, crns)
#
#        crns = self.catalog.crosslisted_with(76097)
#        for e in (76093, 76098, 76099):
#            self.assertIn(e, crns)
#
#        crns = self.catalog.crosslisted_with(76101)
#        self.assertEquals(crns, (76106,))

    def test_get_year(self):
        self.assertEquals(self.catalog.year, 2012)

    def test_get_semester(self):
        self.assertEquals(self.catalog.semester, 'Spring')

    def get_section_by_crn(self, crn):
        for course in self.catalog.get_courses():
            for section in course.sections:
                if crn == section.crn:
                    return section
        raise KeyError("Section with CRN of " + str(crn) + " does not exist.")

    def test_time_period_from_11_to_1pm(self):
        "When period lists 11 - 1:50pm..."
        section = self.get_section_by_crn(96688)
        expected_section = models.Section(
            crn=96688, num='03', taken=0, total=23, notes=[],
            periods=[models.Period(
                type='LAB', instructor='Morris',
                start=1100, end=1250,
                location='', int_days=(3,)
            )]
        )
        self.assertSectionEquals(section, expected_section)

    def test_get_ta_training_seminar_with_no_time_periods(self):
        course = self.catalog.find_courses('TA TRAINING SEMINAR')[0]
        expected_course = models.Course(
            'TA TRAINING SEMINAR', 'ADMN', num=6800, credmin=0, credmax=0,
            grade_type='Satisfactory/Unsatisfactory',
            sections=[models.Section(
                crn=97989, num='01', taken=0, total=200,
                periods=[models.Period(
                    type='LEC', instructor='Gornic',
                    start='** TBA **', end='** TBA **',
                    location='', int_days=(),
                )],
                notes=[]
            )]
        )
        self.assertCourseEquals(course, expected_course)

    def test_get_intro_to_hci(self):
        course = self.catalog.find_courses('INTRODUCTION TO HCI')[0]
        expected_course = models.Course(
            'INTRODUCTION TO HCI', 'ITWS', num=2210, credmin=4, credmax=4,
            grade_type='', sections=[models.Section(
                crn=97772, num='01', taken=0, total=40,
                notes=('RESTRICTED TO COMM, EMAC, ITWS MAJORS',),
                periods=[models.Period(
                    type='LEC', instructor='Grice',
                    start=1200, end=1350, location='',
                    int_days=(0, 3),
                )],
            )]
        )
        self.assertCourseEquals(course, expected_course)

    def test_creative_writing_poetry(self):
        # Bug: this class was not detected by the parser
        course = self.catalog.find_courses('CREATIVE WRITING: POETRY')[0]
        expected_course = models.Course(
            'CREATIVE WRITING: POETRY', 'LITR', num=2961, credmin=4, credmax=4,
            grade_type='', sections=[models.Section(
                crn=96438, num='01', taken=0, total=22,
                periods=[models.Period(
                    type='LEC', instructor='Gutmann',
                    start=1200, end=1350, location='',
                    int_days=(1, 4),
                )],
                notes=['COMMUNICATION INTENSIVE'],
            )]
        )
        self.assertCourseEquals(course, expected_course)

    def test_crosslisted_courses(self):
        course = self.catalog.find_courses('INTRO TO COG NEUROSCIENCE')[1]
        expected_course = models.Course(
            'INTRO TO COG NEUROSCIENCE', 'PSYC', num=4968, credmin=4, credmax=4,
            grade_type='', sections=[models.Section(
                crn=98083, num='01', taken=0, total=10,
                periods=[models.Period(
                    type='LEC', instructor='Walf',
                    start=1400, end=1550, location='',
                    int_days=(1, 4),
                )],
                notes=['MEETS WITH COGS 4967']
            )]
        )
        self.assertCourseEquals(course, expected_course)

    def test_get_psych_course_count(self):
        courses = self.catalog.find_courses('PSYC')
        expected_names = (
            'GENERAL PSYCHOLOGY',
            'CRITICAL THINKING',
            'INTRO. TO COGNITIVE SCIEN',
            'HUMAN FACTORS IN DESIGN',
            'SOCIAL PSYCHOLOGY',
            'THINKING',
            'DEMOCRACY: SOCIAL VS POLI',
            'HEROES OF THE HUDSON VALL',
            'ANARCHISM: ETHICAL SOCIET',
            'MOTIVATION & PERFORMANCE',
            'PROFESSIONAL DEVELOPMENT',
            'ORGANIZATIONAL PSYCHOLOGY',
            'ADV EXPER METHODS & STATI',
            'BEHAVIORAL NEUROSCIENCE',
            'COGNITIVE PSYCHOLOGY',
            'PERSONALITY',
            'LEARNING',
            'DRUGS SOCIETY & BEHAVIOR',
            'ABNORMAL PSYCHOLOGY',
            'FORENSIC PSYCHOLOGY',
            'PSYCHOPHARM & BEH TOXICO',
            'SPORTS PSYCHOLOGY SEMINAR',
            'READINGS IN PSYC',
            'ADVANCED TOPICS IN MOTIVA',
            'COG SCI & ECONOMICS',
            'INTRO TO COG NEUROSCIENCE',
            'STRESS & THE BRAIN',
            'UNDERGRADUATE THESIS',
        )
        self.assertEqual(set(c.name for c in courses), set(expected_names))

    def test_courses_with_multiple_periods(self):
        course = self.catalog.find_courses('DATA STRUCTURES')[0]
        print course.sections[0].periods
        expected_course = models.Course(
            'DATA STRUCTURES', 'CSCI', num=1200, credmin=4, credmax=4,
            grade_type='', sections=[
                models.Section(
                    crn=95551, num='01', taken=0, total=33, notes=[],
                    periods=[
                        models.Period(
                            type='LEC', instructor='Stewart',
                            start=1200, end=1350, location='',
                            int_days=(0, 3)
                        ),
                        models.Period(
                            type='LAB', instructor='Staff',
                            start=1000, end=1150, location='',
                            int_days=(2,)
                        ),
                    ]
                ),
                models.Section(
                    crn=95552, num='02', taken=0, total=33, notes=[],
                    periods=[
                        models.Period(
                            type='LEC', instructor='Stewart',
                            start=1200, end=1350, location='',
                            int_days=(0, 3)
                        ),
                        models.Period(
                            type='LAB', instructor='Staff',
                            start=1000, end=1150, location='',
                            int_days=(2,)
                        ),
                    ]
                ),
                models.Section(
                    crn=95553, num='03', taken=0, total=33, notes=[],
                    periods=[
                        models.Period(
                            type='LEC', instructor='Stewart',
                            start=1200, end=1350, location='',
                            int_days=(0, 3)
                        ),
                        models.Period(
                            type='LAB', instructor='Staff',
                            start=1200, end=1350, location='',
                            int_days=(2,)
                        ),
                    ]
                ),
                models.Section(
                    crn=95554, num='04', taken=0, total=33, notes=[],
                    periods=[
                        models.Period(
                            type='LEC', instructor='Stewart',
                            start=1200, end=1350, location='',
                            int_days=(0, 3)
                        ),
                        models.Period(
                            type='LAB', instructor='Staff',
                            start=1400, end=1550, location='',
                            int_days=(2,)
                        ),
                    ]
                ),
                models.Section(
                    crn=95555, num='05', taken=0, total=33, notes=[],
                    periods=[
                        models.Period(
                            type='LEC', instructor='Stewart',
                            start=1200, end=1350, location='',
                            int_days=(0, 3)
                        ),
                        models.Period(
                            type='LAB', instructor='Staff',
                            start=1600, end=1750, location='',
                            int_days=(2,)
                        ),
                    ]
                ),
                models.Section(
                    crn=95556, num='06', taken=0, total=33, notes=[],
                    periods=[
                        models.Period(
                            type='LEC', instructor='Stewart',
                            start=1200, end=1350, location='',
                            int_days=(0, 3)
                        ),
                        models.Period(
                            type='LAB', instructor='Staff',
                            start=1800, end=1950, location='',
                            int_days=(2,)
                        ),
                    ]
                ),
                models.Section(
                    crn=95557, num='07', taken=0, total=33, notes=[],
                    periods=[
                        models.Period(
                            type='LEC', instructor='Stewart',
                            start=1200, end=1350, location='',
                            int_days=(0, 3)
                        ),
                        models.Period(
                            type='LAB', instructor='Staff',
                            start=1200, end=1350, location='',
                            int_days=(2,)
                        ),
                    ]
                ),
            ]
        )
        self.assertCourseEquals(course, expected_course)

    def test_get_intro_to_phil_with_three_sections(self):
        course = self.catalog.find_courses('INTRO TO PHILOSOPHY')[0]
        expected_course = models.Course(
            'INTRO TO PHILOSOPHY', 'PHIL', num=1110, credmin=4, credmax=4,
            grade_type='', sections=[
                models.Section(
                    crn=96407, num='01', taken=0, total=35,
                    periods=[models.Period(
                        type='LEC', instructor='Fahey',
                        start=1000, end=1150, location='',
                        int_days=(0, 3)
                    )],
                    notes=[]
                ),
                models.Section(
                    crn=95481, num='02', taken=0, total=35,
                    periods=[models.Period(
                        type='LEC', instructor='Carcasole',
                        start=1000, end=1150, location='',
                        int_days=(1, 4)
                    )],
                    notes=[]
                ),
                models.Section(
                    crn=98420, num='03', taken=0, total=27,
                    periods=[models.Period(
                        type='LEC', instructor='Thero',
                        start=1200, end=1350, location='',
                        int_days=(0, 3)
                    )],
                    notes=[]
                )
            ],
        )

        self.assertCourseEquals(course, expected_course)

    def test_digital_control_systems(self):
        course = self.catalog.find_courses('DIGITAL CONTROL SYSTEMS')[0]
        expected_course = models.Course(
            'DIGITAL CONTROL SYSTEMS', 'ECSE', num=4510, credmin=3, credmax=3,
            grade_type='', sections=[
                models.Section(
                    crn=95474, num='01', taken=0, total=40,
                    periods=[models.Period(
                        type='LEC', instructor='Chow',
                        start=1200, end=1320, location='',
                        int_days=(1, 4)
                    )],
                    notes=[]
                )
            ]
        )

        self.assertCourseEquals(course, expected_course)

    def test_digital_communications(self):
        course = self.catalog.find_courses('DIGITAL COMMUNICATIONS')[0]
        expected_course = models.Course(
            'DIGITAL COMMUNICATIONS', 'ECSE', num=4560, credmin=3, credmax=3,
            grade_type='', sections=[
                models.Section(
                    crn=97975, num='01', taken=0, total=40,
                    periods=[models.Period(
                        type='LEC', instructor='Saulnier',
                        start=1600, end=1720, location='',
                        int_days=(0, 3)
                    )],
                    notes=['MEETS WITH ECSE 6560']
                )
            ]
        )

        self.assertCourseEquals(course, expected_course)


if __name__ == '__main__':
    unittest.main()



import unittest

import rpi_courses.scheduler as s
from rpi_courses.parser import CourseCatalog
from pyconstraints import Problem

from constants import XML_SCHEDULE_TEST_FILE, XML_SCHEDULE_TEST_CONFLICT_FILE


with open(XML_SCHEDULE_TEST_FILE) as f:
    SCHEDULE_CONTENTS = f.read()

with open(XML_SCHEDULE_TEST_CONFLICT_FILE) as f:
    CONFLICT_CONTENTS = f.read()


class TestTimeRange(unittest.TestCase):
    "Time range of [1200, 1300]"
    def setUp(self):
        self.tr = s.TimeRange(start=1200, end=1300, dow=(0,))

    def assertIn(self, item, collection):
        if hasattr(super(TestTimeRange, self), 'assertIn'):
            super(TestTimeRange, self).assertIn(item, collection)
        else:
            assert item in collection

    def assertNotIn(self, item, collection):
        if hasattr(super(TestTimeRange, self), 'assertNotIn'):
            super(TestTimeRange, self).assertNotIn(item, collection)
        else:
            assert item not in collection

    def test_time_range_does_contain(self):
        "should contain time range of [1100, 1250]."
        self.assertIn(((0,), 1100, 1250), self.tr)

    def test_time_range_does_not_contain(self):
        "should not contain time range of [1100, 1150]."
        self.assertNotIn(((0,), 1100, 1150), self.tr)


class TestMakeSchedules(unittest.TestCase):
    "Scheduler from XML fixture"
    # TODO: make instances themselves instead of using this XML files.
    def setUp(self):
        catalog = CourseCatalog.from_string(SCHEDULE_CONTENTS)
        course_names = (
            'graph theory math',
            'learning',
            'intro to philosophy',
            'software design & doc',
            'multivar calc',
        )
        self.courses = {}
        for n in course_names:
            self.courses[n] = catalog.find_courses(n)[0]

    def assertIn(self, item, collection):
        if hasattr(super(TestMakeSchedules, self), 'assertIn'):
            super(TestMakeSchedules, self).assertIn(item, collection)
        else:
            assert item in collection

    def dump(self, schedules):
        results = []
        for schedule in schedules:
            result = {}
            for course, section in schedule.items():
                result[course.name] = section.num
            results.append(result)
        for result in results:
            from pprint import pprint
            pprint(result.values())

    def test_compute_schedules(self):
        "Should have 36 solutions for all courses."
        result = s.compute_schedules(self.courses.values())
        # RPI scheduler lists 12 schedules but:
        # - MATH 2010 has two sections for all the 12 schedules
        # - Software Design & Docs has 2 sections for 6 of the schedules
        self.assertEqual(len(result), 36)

    def test_schedules_with_no_conflicts(self):
        "Should have 2 solutions for graph theory & intro to philosophy."
        courses = (
            self.courses['graph theory math'], self.courses['intro to philosophy']
        )
        result = s.compute_schedules(courses)
        expected = [{
            courses[0]: courses[0].sections[0],
            courses[1]: courses[1].sections[1],
        }, {
            courses[0]: courses[0].sections[0],
            courses[1]: courses[1].sections[0],
        }]

        self.assertIn(result[0], expected)
        self.assertIn(result[1], expected)
        self.assertNotEqual(result[0], result[1])

    def test_schedules_with_restricting_times_and_no_conflicts(self):
        "Should have 1 solution for graph theory & intro to philosophy with time restriction."
        courses = (
            self.courses['graph theory math'], self.courses['intro to philosophy']
        )
        restrictions = []
        for p in courses[1].sections[0].periods:
            restrictions.append(
                s.TimeRange(*p.time_range, dow=(0,))
            )
        result = s.compute_schedules(courses, restrictions)
        expected = [{
            courses[0]: courses[0].sections[0],
            courses[1]: courses[1].sections[1],
        }]
        self.assertEqual(list(result), expected)


class TestMakeConflictableSchedules(unittest.TestCase):
    # TODO: make instances themselves instead of using this XML files.
    def setUp(self):
        catalog = CourseCatalog.from_string(CONFLICT_CONTENTS)
        course_names = (
            'BEGINNING PROG FOR ENG',
            'COMPUTER SCIENCE I',
        )
        self.courses = {}
        for n in course_names:
            self.courses[n] = catalog.find_courses(n)[0]

    def assertIn(self, item, collection):
        if hasattr(super(TestMakeConflictableSchedules, self), 'assertIn'):
            super(TestMakeConflictableSchedules, self).assertIn(item, collection)
        else:
            assert item in collection

    def test_schedules_with_conflicts(self):
        "Should schedule courses with some conflicts."
        result = s.compute_schedules(self.courses.values())
        c = (
            self.courses['BEGINNING PROG FOR ENG'],
            self.courses['COMPUTER SCIENCE I'],
        )
        expected = []
        for s1 in c[0].sections:
            for s2 in c[1].sections:
                if not s1.conflicts_with(s2):
                    expected.append({c[0]: s1, c[1]: s2})

        self.assertEqual(len(result), len(expected))
        for r in result:
            self.assertIn(r, expected)
            for r2 in result:
                if r != r2:
                    self.assertNotEqual(r, r2)

    def test_schedules_with_conflicts_and_restricted_times(self):
        "Should schedule courses with some conflicts and time restrictions"
        result = s.compute_schedules(self.courses.values())
        c = (
            self.courses['BEGINNING PROG FOR ENG'],
            self.courses['COMPUTER SCIENCE I']
        )

        restrictions = [s.TimeRange(start=1000, end=1100, dow=(0,))]

        expected = []
        for s1 in c[0].sections:
            if restrictions[0].conflicts_with(s1):
                continue
            for s2 in c[1].sections:
                if not s1.conflicts_with(s2):
                    expected.append({c[0]: s1, c[1]: s2})

        self.assertEquals(len(result), len(expected))
        for r in result:
            self.assertIn(r, expected)
            for r2 in result:
                if r != r2:
                    self.assertNotEqual(r, r2)


if __name__ == '__main__':
    unittest.main()



"""features.py - Implements all parsing of the XML file.

All functions related to parsing the XML file are here. To be
automatically imported by the CourseCatalog class, postfix the
function name with '_feature'
"""
import datetime
import re
from rpi_courses.utils import FrozenDict, safeInt
from rpi_courses.config import logger, DEBUG
from rpi_courses.models import CrossListing, Course, Period, Section


def timestamp_feature(catalog, soup):
    """The datetime the xml file was last modified.
    """
    # there's really no "time created", we're using the date the courses are listed for...
    epoch = 1318790434
    catalog.timestamp = int(float(soup.title.text)) + epoch
    catalog.datetime = datetime.datetime.fromtimestamp(catalog.timestamp)
    logger.info('Catalog last updated on %s' % catalog.datetime)

def _text(nodes, sep='\n'):
    sb = []
    for node in nodes:
        sb.append(node.text.strip())
    return sep.join(sb)

RE_SEMESTER_RANGE = re.compile(r'(?P<start_month>[A-Za-z]+) +(?P<start_day>\d+) +- +(?P<end_month>[A-Za-z]+) +(?P<end_day>\d+),? +(?P<year>\d+)')
RE_SEMESTER_URL = re.compile(r'^.+zs(?P<year>\d{4})(?P<month>\d{2}).+$')

def semester_feature(catalog, soup):
    """The year and semester information that this xml file hold courses for.
    """
    raw = _text(soup.findAll('h3')).split('\n')[1]
    print 'raw', raw
    match = RE_SEMESTER_RANGE.match(raw)
    catalog.year = int(match.group('year'))

    #month_mapping = {'Spring': 1, 'Summer': 5, 'Fall': 9}
    month_mapping = {'january': 1, 'may': 5, 'august': 9}
    catalog.month = month_mapping[match.group('start_month').lower()]

    if catalog.url:
        match = RE_SEMESTER_URL.match(catalog.url)
        if match:
            catalog.year = int(match.group('year'))
            catalog.month = int(match.group('month'))

    semester_mapping = {1: 'Spring', 5: 'Summer', 9: 'Fall'}
    catalog.semester = semester_mapping[catalog.month]
    catalog.name = '%s %d' % (catalog.semester, catalog.year)
    logger.info('Catalog type: %s' % catalog.name)


#def crosslisting_feature(catalog, soup):
#    """Parses all the crosslistings. These refer to the similar CRNs,
#    such as a grad & undergrad level course.
#    """
#    listing = {}
#    for elem in soup.coursedb.findAll('crosslisting'):
#        seats = int(elem['seats'])
#        crns = [safeInt(crn.string) for crn in elem.findAll('crn')]
#
#        # we want to refer to the same object to save space
#        cl = CrossListing(crns, seats)
#        for crn in crns:
#            listing[crn] = cl
#    catalog.crosslistings = FrozenDict(listing)
#
#    logger.info('Catalog has %d course crosslistings' % len(catalog.crosslistings))


def course_feature(catalog, soup):
    """Parses all the courses (AKA, the most important part).
    """
    courses = {}
    count = 0
    for course_data in parse_tables(soup):
        c = create_course(course_data)
        count += 1
        courses[str(c)] = c
    catalog.courses = FrozenDict(courses)
    logger.info('Catalog has %d courses (manual: %d)' % (len(courses), count))


# INTERNAL FUNCTIONS


def create_period(period_data):
    return Period(**period_data)


def create_section(section_data):
    data = dict(section_data)
    data['periods'] = tuple(create_period(p) for p in section_data['periods'])
    return Section(**data)


def create_course(course_data):
    data = dict(course_data)
    data['sections'] = tuple(create_section(s) for s in course_data['sections'])
    return Course(**data)


class_days = {
    'M': 0, 'T': 1, 'W': 2, 'R': 3, 'F': 4
}


def extract_period(cells, period, G):
    # possible choices
    # [u'CRN Course-Sec', u'Course Title', u'Class Type', u'Cred Hrs', u'Gr Tp', u'Building/Room'
    # u'Class Days', u'Start Time', u'End Time', u'Instructor', u'Max Enrl', u'Enrl', u'Sts Rmng']
    #
    # use G(cells, NAME)
    #
    # class type
    period['type'] = G(cells, 'Class Type').text
    # class days
    period['int_days'] = list(class_days[x] for x in G(cells, 'Class Days').text if x.strip() != '')
    # start-time & end-time (we need end-time to figure out if it's in the morning or not)

    def is_tba(s):
        return not s or 'TBA' in s.upper()
    period['start'], period['end'] = G(cells, 'Start Time').text, G(cells, 'End Time').text

    if not is_tba(period['start']) and not is_tba(period['end']):

        is_pm = period['end'].upper().endswith('PM')

        def get_time(s):
            return int(s.replace(':', ''))

        period['start'] = get_time(period['start'])
        period['end'] = get_time(period['end'][:-2])

        if is_pm:
            period['start'] += 1200
            period['end'] += 1200
            if period['start'] >= 2400:
                period['start'] -= 1200
            if period['end'] >= 2400:
                period['end'] -= 1200

            # this covers the case of getting 11:00 - 1:50PM
            # we're assuming there's no classes from the evening that go beyond midnight.
            if period['start'] > period['end']:
                period['start'] -= 1200

        period['start'], period['end'] = str(period['start']), str(period['end'])
    # instructor
    node = G(cells, 'Instructor')
    period['instructor'] = node.text.strip() if node else ''
    # location
    node = G(cells, 'Building/Room')
    period['location'] = node.text.strip() if node else ''


def parse_tables(node):
    courses = []
    cache = {}
    last_course = last_section = last_period = None
    rows = node.findAll('tr')

    columns = [None] * len(rows[0].findAll('th'))

    for i, row in enumerate(rows[0].findAll('th')):
        columns[i] = row.text.strip()

    for i, row in enumerate(rows[1].findAll('th')):
        columns[i] += ' ' + row.text.strip() if row.text else ''

    columns = tuple(x.strip() for x in columns)

    print columns
    # possible choices
    # [u'CRN Course-Sec', u'Course Title', u'Class Type', u'Cred Hrs', u'Gr Tp', u'Building/Room'
    # u'Class Days', u'Start Time', u'End Time', u'Instructor', u'Max Enrl', u'Enrl', u'Sts Rmng']

    def G(cells, name):
        try:
            return cells[columns.index(name)]
        except (IndexError, ValueError):
            return None

    for row in rows[2:]:
        course = {'sections': []}
        section = {'notes': set(), 'periods': []}
        period = {}
        cells = row.findAll('td')
        # if we got to a new course / section
        if len(cells) < 2:
            continue
        elif cells[0].text.strip() != '':
            # <crn> <code>-<num>-<sec>
            parts = G(cells, 'CRN Course-Sec').text.split(' ', 1)
            section['crn'] = parts[0]
            course['dept'], course['num'], section['num'] = parts[1].split('-', 2)
            # course name
            course['name'] = G(cells, 'Course Title').text.strip()
            existing_obj = cache.get(course['name'] + course['dept'] + course['num'])
            if existing_obj:
                course = existing_obj

            # credit hours (eg - '1' or '1-6')
            parts = G(cells, 'Cred Hrs').text.split('-', 1)
            course['credmin'], course['credmax'] = parts[0], parts[1] if len(parts) > 1 else parts[0]
            # grade_type
            grade_type = G(cells, 'Gr Tp')
            if grade_type:
                course['grade_type'] = {
                    'SU': 'Satisfactory/Unsatisfactory',
                }.get(grade_type.text, grade_type.text)
            # seats total
            node = G(cells, 'Max Enrl')
            if node:
                section['total'] = int(node.text) if node.text.strip() else 0
            else:
                section['total'] = ''
            # seats taken
            node = G(cells, 'Enrl')
            if node:
                section['taken'] = int(node.text) if node.text.strip() else 0
            else:
                section['taken'] = ''
            # textbook link? could be usedful
            # section['textbook_link'] = cells[12].find('a')['href']
            extract_period(cells, period, G)

            # link up
            section['periods'].append(period)
            course['sections'].append(section)
            cache[course['name'] + course['dept'] + course['num']] = course

            if not existing_obj:
                courses.append(course)

            last_course, last_section, last_period = course, section, period

        elif 'NOTE:' in cells[1].text.strip():  # process note
            course, section = last_course, last_section
            section['notes'].add(cells[2].text.strip())

        else:  # process a new period type
            course, section = last_course, last_section

            period = last_period.copy()
            extract_period(cells, period, G)
            section['periods'].append(period)

    return courses



"""web.py - Handles all the http interaction of getting the course
catalog data. This isn't your web.py web dev framework!
"""
import urllib2
import datetime
import tempfile
import pyPdf
from contextlib import closing

from BeautifulSoup import BeautifulSoup

from config import ROCS_URL, SIS_URL, COMM_URL


def get(url, last_modified=None):
    """Performs a get request to a given url. Returns an empty str on error.
    """
    try:
        with closing(urllib2.urlopen(url)) as page:
            if last_modified is not None:
                last_mod = dateutil.parser.parse(dict(page.info())['last-modified'])
                if last_mod <= last_modified:
                    return ""
            return page.read()
    except urllib2.URLError:
        return ""


def list_sis_files_for_date(date=None, url_base=SIS_URL):
    date = date or datetime.datetime.now()
    format = '%szs%.4d%.2d.htm'
    base = []
    months = (1, 5, 9)
    prev_m = None
    for m in months:
        if m >= date.month:
            base.append(format % (url_base, date.year, m))
            if prev_m and prev_m < date.month:
                base.append(format % (url_base, date.year, prev_m))
        prev_m = m
    if not base:
        base.append(format % (url_base, date.year + 1, 1))
    return base


def list_sis_files(url_base=SIS_URL):
    date = datetime.date(year=2011, month=1, day=1)
    today = datetime.date.today()
    urls = []
    while date.year <= today.year + 1:
        urls.extend(list_sis_files_for_date(date, url_base=url_base))
        date = datetime.date(year=date.year + 1, month=1, day=1)
    return urls


def list_rocs_files(url=ROCS_URL):
    """Gets the contents of the given url.
    """
    soup = BeautifulSoup(get(url))
    if not url.endswith('/'):
        url += '/'
    files = []
    for elem in soup.findAll('a'):
        if elem['href'].startswith('?'):
            continue
        if elem.string.lower() == 'parent directory':
            continue
        files.append(url + elem['href'])
    return files


def is_xml(filename):
    "Returns True if the filename ends in an xml file extension."
    return filename.strip().endswith('.xml')


def list_rocs_xml_files(url=ROCS_URL):
    "Gets all the xml files."
    return list(filter(is_xml, list_rocs_files(url)))


def get_comm_file(date, base_url=COMM_URL):
    format = '%.4d.pdf'
    if date.month == 9:
        url = base_url + "Fall" + str(format % (date.year))
    else:
        url = base_url + "Spring" + str(format % (date.year))

    req = urllib2.Request(url)
    print "Getting communication intensive list from: " + url

    full_text = ""
    tmp = None
    try:
        f = urllib2.urlopen(req)
        temp = tempfile.NamedTemporaryFile()
        temp.write(f.read())
        temp.seek(0)
        with open(temp.name, 'rb'):
            pdf = pyPdf.PdfFileReader(open(temp.name, 'rb'))
            for page in pdf.pages:
                full_text += page.extractText()

    except urllib2.HTTPError, e:
        print "HTTP Error:", e.code, url
    except urllib2.URLError, e:
        print "URL Error:", e.reason, url
    finally:
        if temp:
            temp.close()
    return full_text.strip()



from rpi_courses.parser import CourseCatalog
from rpi_courses import models
from test.rpi_courses.utils import TestCaseForModel

from mock import Mock

import os
import unittest
from time import mktime
from constants import XML_FILE, XML_PARSER_FILE


with open(XML_PARSER_FILE) as f:
    CONTENTS = f.read()


class TestCatalog(TestCaseForModel):
    def setUp(self):
        self.catalog = CourseCatalog.from_string(CONTENTS)

    def test_crosslisting(self):
        crns = self.catalog.crosslisted_with(76093)
        for e in (76097, 76098, 76099):
            self.assertIn(e, crns)

        crns = self.catalog.crosslisted_with(76097)
        for e in (76093, 76098, 76099):
            self.assertIn(e, crns)

        crns = self.catalog.crosslisted_with(76101)
        self.assertEquals(crns, (76106,))

    def test_get_timestamp(self):
        expected_timestamp = 1290023154
        timestamp = int(mktime(self.catalog.datetime.timetuple()))

        self.assertEquals(timestamp, expected_timestamp)

    def test_get_year(self):
        self.assertEquals(self.catalog.year, 2011)

    def test_get_semester(self):
        self.assertEquals(self.catalog.semester, 'Spring')

    def test_get_ta_training_seminar_with_no_time_periods(self):
        course = self.catalog.find_courses('TA TRAINING SEMINAR')[0]
        expected_course = models.Course(
            'TA TRAINING SEMINAR', 'ADMN', num=6800, credmin=0, credmax=0,
            grade_type='Satisfactory/Unsatisfactory',
            sections=[models.Section(
                crn=76285, num='01', taken=0, total=200,
                periods=[models.Period(
                    type='LEC', instructor='Gornic',
                    start='** TBA **', end='** TBA **',
                    location=' ', int_days=(),
                )],
                notes=[
                    'REQUIRED FOR ALL NEW OR CURRENT STUDENTS WHO WILL TA',
                    'MEETING ON THUR, 1/20 IN PM AND ALL DAY 1/21'
                ]
            )]
        )
        self.assertCourseEquals(course, expected_course)

    def test_get_intro_to_hci(self):
        course = self.catalog.find_courses('INTRODUCTION TO HCI')[0]
        expected_course = models.Course(
            'INTRODUCTION TO HCI', 'ITWS', num=2210, credmin=4, credmax=4,
            grade_type='', sections=[models.Section(
                crn=76041, num='01', taken=26, total=40,
                periods=[models.Period(
                    type='LEC', instructor='Grice',
                    start=1200, end=1350, location='LALLY 102',
                    int_days=(0, 3),
                )],
                notes=['RESTRICTED TO COMM, ITWS & EMAC MAJORS']
            )]
        )
        self.assertCourseEquals(course, expected_course)

    def test_get_intro_to_phil_with_two_sections(self):
        course = self.catalog.find_courses('INTRO TO PHILOSOPHY')[0]
        expected_course = models.Course(
            'INTRO TO PHILOSOPHY', 'PHIL', num=1110, credmin=4, credmax=4,
            grade_type='', sections=[
                models.Section(
                    crn=74447, num='01', taken=35, total=35,
                    periods=[models.Period(
                        type='LEC', instructor='Fahey',
                        start=1000, end=1150, location='CARNEG 206',
                        int_days=(0, 3)
                    )],
                    notes=[]
                ),
                models.Section(
                    crn=73498, num='02', taken=35, total=35,
                    periods=[models.Period(
                        type='LEC', instructor='Carcasole',
                        start=1000, end=1150, location='CARNEG 101',
                        int_days=(1, 4)
                    )],
                    notes=[]
                )
            ],
        )

        self.assertCourseEquals(course, expected_course)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('README.rst') as handle:
    README = handle.read()

setup(
    name='RPICourses',
    description='A data scraping library to read course data from RPI.',
    license='MIT',
    version='1.0.4',
    author='Jeff Hui',
    author_email='jeff@jeffhui.net',
    url='https://github.com/jeffh/rpi_courses',
    packages=find_packages(),
    requires=[
        'BeautifulSoup',
        'pyPdf',
        'pyconstraints',
    ],
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable', # I wish
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.6',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries',
    ],
)



import unittest
import urllib2
from rpi_courses import web
from mock import patch
from constants import HTML


class TestListUnits(unittest.TestCase):
    def setUp(self):
        self.url = "http://sis.rpi.edu/reg/rocs/"
        self.expected_files = map(lambda x: self.url + x, [
            '201001.xml',
            '201005.xml',
            '201009.xml',
            '201009.xml_old',
            '201101.xml',
            'testrocs.txt',
        ])

    @patch.object(web, 'get')
    def test_list_rocs_files(self, mock):
        """Lists files to read from a given url page.
        Expects an apache file listing page.
        """
        mock.return_value = HTML

        files = web.list_rocs_files(self.url)

        mock.assert_called_with(self.url)

        self.assertEqual(files, self.expected_files)

    @patch.object(web, 'get')
    def test_list_rocs_xml_files(self, mock):
        """Lists all the files ending with '.xml'.
        Expects an apache file listing page.
        """
        mock.return_value = HTML
        files = web.list_rocs_xml_files(self.url)

        mock.assert_called()

        self.expected_files.remove(self.url + '201009.xml_old')
        self.expected_files.remove(self.url + 'testrocs.txt')

        self.assertEqual(files, self.expected_files)

if __name__ == '__main__':
    unittest.main()



from pyconstraints import Problem, is_nil, BruteForceSolver


__all__ = ['compute_schedules', 'TimeRange', 'Scheduler']


class TimeRange(object):
    "Represents a time range to be restricted."
    def __init__(self, start, end, dow):
        self.start = start
        self.end = end
        self.days_of_week = dow

    def __repr__(self):
        return "<TimeRange: %r to %r on %r>" % (
            self.start, self.end, self.days_of_week
        )

    def days_conflict(self, days):
        print self.days_of_week, days
        for day in self.days_of_week:
            if day in days:
                return True
        return False

    def __contains__(self, period):
        days, start, end = period

        return self.days_conflict(days) and (
            self.start <= start <= self.end or
            start <= self.start <= end or
            self.start <= end <= self.end or
            start <= self.end <= end
        )

    def conflicts_with(self, section):
        "Returns True if the given section conflicts with this time range."
        for p in section.periods:
            t = (p.int_days, p.start, p.end)
            if t in self:
                return True
        return False


def section_constraint(section1, section2):
    if is_nil(section1) or is_nil(section2):
        return True
    return not section1.conflicts_with(section2)


class Scheduler(object):
    """High-level API that wraps the course scheduling feature.

    ``free_sections_only``: bool. Determines if the only the available sections should be
                            used when using courses provided. Defaults to True.
    ``problem``: Optional problem instance to provide. If None, the default one is created.

    """
    def __init__(self, free_sections_only=True, problem=None, constraint=None):
        self.p = Problem()
        if problem is not None:
            self.p = problem
        self.free_sections_only = free_sections_only
        self.section_constraint = constraint or section_constraint
        self.clear_excluded_times()

    def clear_excluded_times(self):
        """Clears all previously set excluded times."""
        self._excluded_times = []
        return self

    def exclude_time(self, start, end, days):
        """Added an excluded time by start, end times and the days.

        ``start`` and ``end`` are in military integer times (e.g. - 1200 1430).
        ``days`` is a collection of integers or strings of fully-spelt, lowercased days
                 of the week.
        """
        self._excluded_times.append(TimeRange(start, end, days))
        return self

    def exclude_times(self, *tuples):
        """Adds multiple excluded times by tuple of (start, end, days) or by
        TimeRange instance.

        ``start`` and ``end`` are in military integer times (e.g. - 1200 1430).
        ``days`` is a collection of integers or strings of fully-spelt, lowercased days
                 of the week.
        """
        for item in tuples:
            if isinstance(item, TimeRange):
                self._excluded_times.append(item)
            else:
                self.exclude_time(*item)
        return self

    def find_schedules(self, courses=None, return_generator=False):
        """Returns all the possible course combinations. Assumes no duplicate courses.

        ``return_generator``: If True, returns a generator instead of collection. Generators
            are friendlier to your memory and save computation time if not all solutions are
            used.
        """
        self.p.reset()
        self.create_variables(courses)
        self.create_constraints(courses)
        if return_generator:
            return self.p.iter_solutions()
        return self.p.get_solutions()

    # internal methods -- can be overriden for custom use.
    def get_sections(self, course):
        """Internal use. Returns the sections to use for the solver for a given course.
        """
        return course.available_sections if self.free_sections_only else course.sections

    def time_conflict(self, schedule):
        """Internal use. Determines when the given time range conflicts with the set of
        excluded time ranges.
        """
        if is_nil(schedule):
            return True
        for timerange in self._excluded_times:
            if timerange.conflicts_with(schedule):
                return False
        return True

    def create_variables(self, courses):
        """Internal use. Creates all variables in the problem instance for the given
        courses. If given a dict of {course: sections}, will use the provided sections.
        """
        has_sections = isinstance(courses, dict)
        for course in courses:
            self.p.add_variable(course, courses.get(course, []) if has_sections else self.get_sections(course))

    def create_constraints(self, courses):
        """Internal use. Creates all constraints in the problem instance for the given
        courses.
        """
        for i, course1 in enumerate(courses):
            for j, course2 in enumerate(courses):
                if i <= j:
                    continue
                self.p.add_constraint(self.section_constraint, [course1, course2])
            self.p.add_constraint(self.time_conflict, [course1])


def compute_schedules(courses=None, excluded_times=(), free_sections_only=True, problem=None, return_generator=False, section_constraint=None):
    """
    Returns all possible schedules for the given courses.
    """
    s = Scheduler(free_sections_only, problem, constraint=section_constraint)
    s.exclude_times(*tuple(excluded_times))
    return s.find_schedules(courses, return_generator)
